// -*- C++ -*-
///////////////////////////////////////////////////////////////////////////
// Clocks -- a class for measuring time within a region of program
// execution.
///////////////////////////////////////////////////////////////////////////

#include <sys/time.h>

namespace std {

class Clock {
protected:
#ifdef PARCH_rs6000
    typedef timebasestruct_t Time_val;
    Time_val get_time() {
	Time_val t;
	::read_real_time(&t,TIMEBASE_SZ) ;
	::time_base_to_time(&t,TIMEBASE_SZ) ;
	return t;
    }
    double subtr_times(const Time_val &start, const Time_val &end) {
	unsigned long start_sec = start.tb_high;
	unsigned long start_nsec = start.tb_low;
	unsigned long end_sec = end.tb_high;
	unsigned long end_nsec = end.tb_low;

	signed long sec = (signed long)end_sec - (signed long)start_sec;
	signed long nsec = (signed long) end_nsec - (signed long)start_nsec;

	// To handle wrap-around.
	// Is this really necessary? Probably not, but the example in the man
	// page does, so I probably should too...
	if (nsec < 0) {
	    sec-- ;
	    nsec += NS_PER_SEC ;
	}

	double seconds = (double)nsec/1e9;
	seconds += (double)sec;
	return seconds;
    }
#else // PARCH_rs6000
    typedef struct timeval Time_val;
    Time_val get_time() {
	Time_val t;
	int stat =  gettimeofday(&t, NULL);
	((void)stat);
	assert(stat != -1);
	return t;
    }
    double subtr_times(const Time_val &start, const Time_val &end) {
	long start_sec = start.tv_sec;
	long start_usec = start.tv_usec;
	long end_sec = end.tv_sec;
	long end_usec = end.tv_usec;

	double seconds = (double)(end_usec - start_usec)/1e6;
	seconds += (double)(end_sec - start_sec);
	return seconds;
    }
#endif // PARCH_rs6000

    Time_val start;

public:
    Clock() : start(get_time()) { }
    double elapse() { return subtr_times(start, get_time()); }
};


} // namespace std
