///////////////////////////////////////////////////////////////////////////
// Join implementations
///////////////////////////////////////////////////////////////////////////

#define JOIN(ji, X_ITER, x_begin, x_end, Y_ITER, y_begin, y_end) \
do { \
    typename join_strategy<X_ITER, typename X_ITER::iterator_tag, \
                  Y_ITER, typename Y_ITER::iterator_tag>::t \
         ji((x_begin), (x_end), (y_begin), (y_end)); \
    for ( ; !ji.done(); ji.next()) {
#define END_JOIN \
    } \
} while(false)


// Join results must implement the following methods:
//     bool done();
//     void next();
//     X_ITER first();
//     Y_ITER second();

template<class X_ITER, class X_TAG, class Y_ITER, class Y_TAG>
class join_strategy;

////////////////////////////////////////////////////////////////////////
// Nest and search

template<class X_ITER, class Y_ITER>
class nest_and_search_join_result {
protected:
    X_ITER x_it; const X_ITER x_end;
    Y_ITER y_it; const Y_ITER y_begin; const Y_ITER y_end;
    inline void advance() {
// *fixme*!!!!
// 	while (!done()) {
// 	    if (*y_begin <= *x_it && *x_it <= *y_end) {
// 		y_it = y_begin;
// 		y_it += *x_it - *y_begin;
// 		return;
// 	    }
// 	    x_it++;
// 	}
	y_it = *x_it;
    }
public:
    nest_and_search_join_result(const X_ITER &x_begin, const X_ITER &x_end,
	      const Y_ITER &y_begin, const Y_ITER &y_end)
	: x_it(x_begin), x_end(x_end),
	  y_begin(y_begin), y_end(y_end) {
#ifdef DEBUG
	cout << "nest_and_search_join_result" << endl;
#endif
	advance();
    }
    inline bool done() { return (x_it == x_end); }
    inline void next() { x_it++; advance(); }
    inline X_ITER first() { return x_it; }
    inline Y_ITER second() { return y_it; }
};

////////////////////////////////////////////////////////////////////////
// two-finger

template<class X_ITER, class Y_ITER>
class two_finger_join_result {
protected:
    X_ITER x_it; const X_ITER x_end;
    Y_ITER y_it; const Y_ITER y_end;
    void advance() {
	while (!done()) {
	    if (*x_it < *y_it) x_it++;
	    else if (*x_it > *y_it) y_it++;
	    else return;
	}
    }
public:
    two_finger_join_result(const X_ITER &x_begin, const X_ITER &x_end,
	      const Y_ITER &y_begin, const Y_ITER &y_end)
	: x_it(x_begin), x_end(x_end),
	  y_it(y_begin), y_end(y_end) {
#ifdef DEBUG
	cout << "two_finger_join_result" << endl;
#endif
	advance();
    }
    bool done() { return x_it == x_end || y_it == y_end; }
    void next() { x_it++; y_it++; advance(); }
    X_ITER first() { return x_it; }
    Y_ITER second() { return y_it; }
};

////////////////////////////////////////////////////////////////////////
// two intervals

template<class X_ITER, class Y_ITER>
class interval_join_result {
protected:
    interval_iterator i; interval_iterator ub;
public:
    interval_join_result(const X_ITER &_x_begin,
		       const X_ITER &_x_end,
		       const Y_ITER &_y_begin,
		       const Y_ITER &_y_end) {
#ifdef DEBUG
	cout << "interval_join_result" << endl;
#endif
	if (*_x_begin > *_y_begin) i = _x_begin; else i = _y_begin;
	if (*_x_end < *_y_end) ub = _x_end; else ub = _y_end;
    }
    bool done() { return *i >= *ub; }
    void next() { i++; }
    X_ITER first() { return i; }
    Y_ITER second() { return i; }
};

////////////////////////////////////////////////////////////////////////

#define DEFINE_JOIN_STRATEGY(X_TAG,Y_TAG,STRATEGY) \
template<class X_ITER, class Y_ITER> \
class join_strategy<X_ITER,X_TAG,Y_ITER,Y_TAG> { \
public: \
    typedef STRATEGY<X_ITER,Y_ITER> t; \
}

DEFINE_JOIN_STRATEGY(increasing_iterator_tag,interval_iterator_tag,
		     nest_and_search_join_result);
DEFINE_JOIN_STRATEGY(ordered_iterator_tag,interval_iterator_tag,
		     nest_and_search_join_result);
DEFINE_JOIN_STRATEGY(offset_iterator_tag,interval_iterator_tag,
		     nest_and_search_join_result);
DEFINE_JOIN_STRATEGY(increasing_iterator_tag,increasing_iterator_tag,
		     two_finger_join_result);
DEFINE_JOIN_STRATEGY(interval_iterator_tag,interval_iterator_tag,
		     interval_join_result);
