#include <iostream.h>

#include <assert.h>
#include <utility>
#include <map>
#include <vector>
#include <ctype.h>
#include <sys/time.h>

#ifdef DEBUG
#define bsp_abort (assert(false))
#else
#define bsp_abort (abort())
#endif

#ifdef __GNUC__
#define bsp_restrict __restrict
#endif

#ifdef __KCC
#define bsp_restrict restrict
#endif

#ifndef bsp_restrict
#define bsp_restrict
#endif

#include "Clocks.hh"

namespace std {

#include "numbers.hh"
#include "tuples.hh"
#include "iterators.hh"
#include "joins.hh"
#include "arrays.hh"
#include "terms.hh"

#include "sparsev.hh"
#include "dense.hh"

///////////////////////////////////////////////////////////////////////////

template<class X_CLASS, class X_ITER, 
         class Y_CLASS, class Y_ITER, 
         class BASE>
void dot_body(
    const term_nesting< X_CLASS, _1, X_ITER, term_scalar<BASE> >&x,
    const term_nesting< Y_CLASS, _1, Y_ITER, term_scalar<BASE> >&y,
    BASE & bsp_restrict result)
{
#ifdef DEBUG
    cout << "dot_body(term_nesting< ..." << endl;
#endif
    assert(y.size(0) == x.size(0));

    result = 0.0;
    JOIN(ji, X_ITER, x.begin(), x.end(), Y_ITER, y.begin(), y.end()) {
	result += x.subterm(ji.first()) * y.subterm(ji.second());
    } END_JOIN;
}

template<class BASE>
// inline 
void dot_body(
    const sparse_vector<BASE> &x,
    const dense_vector<BASE> &y,
    BASE & bsp_restrict result)
{
#ifdef DEBUG
    cout << "dot_body(sparse_vector,dense_vector!" << endl;
#endif
    assert(y.size(0) == x.size(0));

    int * bsp_restrict x_indices = x.indices;
    BASE * bsp_restrict x_storage = x.storage;
    int x_lb = x.lb;
    int x_ub = x.ub;
    BASE * bsp_restrict yy = y.storage; yy += y.lb;

    result = 0.0;
    for (int ii=x_lb; ii<x_ub; ii++)
	result += x_storage[ii] * yy[x_indices[ii]];
}

///////////////////////////////////////////////////////////////////////////

int main(int argc,char **argv) 
{
    (void)argc; (void)argv;

    int n = 10000;

    int x_nz = n;
    int *x_indices = new int[n];
    double *x_values = new double[n];
    {for (int i=0; i<n; i++) {
	x_indices[i] = i;
	x_values[i] = 1.;
    }}
    sparse_vector<double> x_fast(n, x_nz, x_indices, x_values);
    term_nesting< sparse_vector<double>, _1, sparse_vector_iterator<double>, 
	          term_scalar<double> >
	& bsp_restrict x_slow = x_fast;

    double *y_values = new double[n];
    {for (int i=0; i<n; i++)
	y_values[i] = i;}
    dense_vector<double> y(n, y_values);

#ifdef DEBUG
    int steps = 1;
#else
    int steps = 1000;
#endif
    long flops = 2 * x_nz * steps;

    double slow_result;
    Clock c_slow;
    {for (int i=0; i<steps; i++)
	dot_body(x_slow, y, slow_result);}
    double slow_elapse_time = c_slow.elapse();

    double slow_mflops = (double)flops / (slow_elapse_time * 1e6);
    cout << "slow mflops = " << slow_mflops << endl;

    double fast_result;
    Clock c_fast;
    {for (int i=0; i<steps; i++)
	dot_body(x_fast, y, fast_result);}
    double fast_elapse_time = c_fast.elapse();


    double fast_mflops = (double)flops / (fast_elapse_time * 1e6);
    cout << "fast mflops = " << fast_mflops << endl;

    if (slow_result != fast_result) {
	cout << "failed." << endl;
	abort();
    }
    cout << "fast/slow = " << fast_mflops/slow_mflops << endl;

    return 0;
}

};
