template<class BASE> class sparse_vector_iterator;

template<class BASE>
class sparse_vector
    : public term_nesting< sparse_vector<BASE>, _1, sparse_vector_iterator<BASE>, term_scalar<BASE> >
{
protected:
    typedef term_nesting< sparse_vector<BASE>, _1, sparse_vector_iterator<BASE>, term_scalar<BASE> >
    base_term_type;
public:
    typedef typename base_term_type::iterator_type iterator_type;
    typedef typename base_term_type::subterm_type subterm_type;
    typedef typename base_term_type::key_type key_type;
public:
    int * bsp_restrict indices;
    BASE * bsp_restrict storage;
    int lb; int ub; // ub is not inclusive.
    sparse_vector()
 	: base_term_type(),
          indices(0), storage(0),
	  lb(-1), ub(-1) { }
    sparse_vector(int n, int nz, int *indices, BASE *storage, int lb = 0)
	: base_term_type(tuple<_1>(n)),
	  indices(indices), storage(storage),
	  lb(lb), ub(lb+nz)  { }
    ~sparse_vector() { }
    iterator_type v_tn_begin() const
	{ return sparse_vector_iterator<BASE>(indices, storage, lb); }
    iterator_type v_tn_end() const
	{ return sparse_vector_iterator<BASE>(indices, storage, ub); }
    subterm_type v_tn_subterm(iterator_type it) {
	return storage[it.jj]; }
};

template<class BASE>
class sparse_vector_iterator
{
    friend class sparse_vector<BASE>;
public:
    IS_AN_OFFSET_ITERATOR(sparse_vector_iterator<BASE>,_1);
protected:
    int * bsp_restrict indices; BASE * bsp_restrict storage;
    int jj;
public:
    sparse_vector_iterator(int *indices = 0, BASE *storage = 0, int jj = -1) :
	indices(indices), storage(storage), jj(jj) { }
    const key_type v_deref() const { return indices[jj]; }
    bool v_equal(const sparse_vector_iterator<BASE> &y) const
	{ return jj == y.jj; }
    void v_set(const sparse_vector_iterator<BASE> &y)
	{ jj = y.jj; }
    int v_distance(const sparse_vector_iterator<BASE> &y)
	{ return jj - y.jj; }
    void v_incr_delta(int d) { jj += d; }
    void v_decr_delta(int d) { jj -= d; }
public:
    OFFSET_ITERATOR_METHODS(sparse_vector_iterator<BASE>,_1);
};
