///////////////////////////////////////////////////////////////////////////

#define tuple_body(N) \
    typedef N dimen; \
    int get(int p) const { \
	return const_cast<tuple<N>&>(*this).ref(p); \
    } \
    void set(int p, int v) { \
	this->ref(p) = v; \
    } \
    int compare(const tuple<N> &x) const { \
	for (int i=0; i < dimen::as_int; i++) \
	    if (get(i) != x.get(i)) return get(i) - x.get(i); \
	return 0; \
    } \
    bool operator ==(const tuple<N> &x) const \
	{ return (compare(x) == 0);  } \
    bool operator !=(const tuple<N> &x) const \
	{ return (compare(x) != 0);  } \
    bool operator <(const tuple<N> &x) const \
	{ return (compare(x) < 0);  } \
 \
    void operator+=(const tuple<N> &x) { \
	for (int i=0; i < dimen::as_int; i++) \
	    ref(i) += x.get(i); \
    } \
	 \
    void operator-=(const tuple<N> &x) { \
	for (int i=0; i < dimen::as_int; i++) \
	    ref(i) -= x.get(i); \
    }

template<class N> class tuple;

///////////////////////////////////////////////////////////////////////////

class tuple<_0> {
public:
    int &ref(int p) {
	switch (p) {
	default: bsp_abort;
	}
    }
    tuple_body(_0);
};

///////////////////////////////////////////////////////////////////////////

class tuple<_1> {
public:
    int i;
    tuple() : i(-1) { }
    tuple(int i) : i(i) { }
    tuple(tuple<_0>, int i) : i(i) { }
    int first() { return i; }
    tuple<_0> rest() { return tuple<_0>(); }
    int &ref(int p) {
	switch (p) {
	case 0: return i;
	default: bsp_abort;
	}
    }
    operator int () const { return i; }
    tuple_body(_1);
};

///////////////////////////////////////////////////////////////////////////

class tuple<_2> {
public:
    int i, j;
    tuple() : i(-1), j(-1) { }
    tuple(int i, int j) : i(i), j(j) { }
    tuple(tuple<_1> x, int j) : i(x.i), j(j) { }
    int first() { return i; }
    tuple<_1> rest() { return tuple<_1>(j); }
    int &ref(int p) {
	switch (p) {
	case 0: return i;
	case 1: return j;
	default: bsp_abort;
	}
    }
    tuple_body(_2);
};

///////////////////////////////////////////////////////////////////////////

class tuple<_3> {
public:
    int i, j, k;
    tuple() : i(-1), j(-1), k(-1) { }
    tuple(int i, int j, int k) : i(i), j(j), k(k) { }
    tuple(tuple<_2> x, int k) : i(x.i), j(x.j), k(k) { }
    int first() { return i; }
    tuple<_2> rest() { return tuple<_2>(j,k); }
    int &ref(int p) {
	switch (p) {
	case 0: return i;
	case 1: return j;
	case 2: return k;
	default: bsp_abort;
	}
    }
    tuple_body(_3);
};

///////////////////////////////////////////////////////////////////////////

class tuple<_4> {
public:
    int i, j, k, l;
    tuple() : i(-1), j(-1), k(-1), l(-1) { }
    tuple(int i, int j, int k, int l) : i(i), j(j), k(k), l(l) { }
    tuple(tuple<_3> x, int l) : i(x.i), j(x.j), k(x.k), l(l) { }
    int first() { return i; }
    tuple<_3> rest() { return tuple<_3>(j,k,l); }
    int &ref(int p) {
	switch (p) {
	case 0: return i;
	case 1: return j;
	case 2: return k;
	case 3: return l;
	default: bsp_abort;
	}
    }
    tuple_body(_4);
};

///////////////////////////////////////////////////////////////////////////

template<class M, class N>
inline void tuple_head(const tuple<M> &x, tuple<N> &result)
{
    assert(M::as_int >= N::as_int);
    for (int i=0; i<N::as_int; i++)
	result.set(i,x.get(i));
}

template<class M, class N>
inline void tuple_tail(const tuple<M> &x, tuple<N> &result)
{
    assert(M::as_int >= N::as_int);
    for (int i=0; i<N::as_int; i++)
	result.set(i,x.get(M::as_int-N::as_int+i));
}

template<class N, class M>
inline tuple< _plus(N,M) > tuple_concat(const tuple<N> &x, const tuple<M> &y)
{
    tuple< _plus(N,M) > result;
    for (int i=0; i<N::as_int; i++)
	result.set(i,x.get(i));
    for (int i=0; i<M::as_int; i++)
	result.set(N::as_int+i,y.get(i));
    return result;
}

///////////////////////////////////////////////////////////////////////////
